@@ -18,6 +18,7 @@ extern "C" {
 #include <algorithm>
 #include <fstream>
 #include <map>
+#include <cassert>
 #include <set>
 #include <vector>
 
@@ -311,7 +312,7 @@ alias_expansion_builder
 		     vector<derived_probe *> & finished_results)
   {
     // We're going to build a new probe and wrap it up in an
-    // alias_expansion_probe so that the expansion loop recognizer it as
+    // alias_expansion_probe so that the expansion loop recognizes it as
     // such and re-expands its expansion.
     
     probe * n = new probe();
@@ -520,8 +521,7 @@ struct mutated_var_collector
 	if (vd)
 	  mutated_vars->insert (vd);
       }
-    e->left->visit (this);
-    e->right->visit (this);
+    traversing_visitor::visit_assignment(e);
   }
 
   void visit_arrayindex (arrayindex *e)
@@ -534,6 +534,7 @@ struct mutated_var_collector
 	else
 	  throw semantic_error("Assignment to read-only histogram bucket", e->tok);
       }
+    traversing_visitor::visit_arrayindex (e);
   }
 };
 
@@ -570,6 +571,7 @@ struct no_var_mutation_during_iteration_check
 	      }
 	  }
       }
+    traversing_visitor::visit_arrayindex (e);
   }
 
   void visit_functioncall (functioncall* e)
@@ -591,8 +593,7 @@ struct no_var_mutation_during_iteration_check
 	  }
       }
 
-    for (unsigned i=0; i<e->args.size(); i++)
-      e->args[i]->visit (this);
+    traversing_visitor::visit_functioncall (e);
   }
 
   void visit_foreach_loop(foreach_loop* s)
@@ -602,9 +603,7 @@ struct no_var_mutation_during_iteration_check
     if (vd)
       vars_being_iterated.push_back (vd);
     
-    for (unsigned i=0; i<s->indexes.size(); i++)
-      s->indexes[i]->visit (this);
-    s->block->visit (this);
+    traversing_visitor::visit_foreach_loop (s);
 
     if (vd)
       vars_being_iterated.pop_back();
@@ -1594,6 +1593,11 @@ typeresolution_info::visit_arrayindex (arrayindex* e)
       if (e->indexes[0]->type != pe_long)
 	unresolved (e->tok);
       hist->visit (this);
+      if (e->type != pe_long)
+	{
+	  e->type = pe_long;
+	  resolved (e->tok, pe_long);
+	}
       return;
     }
 
@@ -1914,7 +1918,12 @@ typeresolution_info::visit_print_format (print_format* e)
 {
   size_t unresolved_args = 0;
 
-  if (e->print_with_format)
+  if (e->hist)
+    {
+      e->hist->visit(this);
+    }
+
+  else if (e->print_with_format)
     {
       // If there's a format string, we can do both inference *and*
       // checking.
