@@ -28,19 +28,24 @@ struct location
   void *fail_arg;
   void (*emit_address) (void *fail_arg, struct obstack *, Dwarf_Addr);
 
-  const Dwarf_Loc *ops;
+  const Dwarf_Op *ops;
   size_t nops;
 
   Dwarf_Word byte_size;
 
-  enum { loc_address, loc_register, loc_noncontiguous, loc_final } type;
+  enum
+    {
+      loc_address, loc_register, loc_noncontiguous,
+      loc_decl, loc_fragment, loc_final
+    } type;
+  struct location *frame_base;
   union
   {
-    struct			/* loc_address or loc_final */
+    struct			/* loc_address, loc_fragment, loc_final */
     {
+      const char *declare;	/* Temporary that needs declared.  */
       char *program;		/* C fragment, leaves address in s0.  */
       unsigned int stack_depth;	/* Temporaries "s0..<N>" used by it.  */
-      struct location *frame_base;
       bool used_deref;		/* Program uses "deref" macro.  */
     } address;
     unsigned int regno;		/* loc_register */
@@ -55,6 +60,8 @@ alloc_location (struct obstack *pool, struct location *origin)
   loc->fail = origin->fail;
   loc->fail_arg = origin->fail_arg;
   loc->emit_address = origin->emit_address;
+  loc->byte_size = 0;
+  loc->frame_base = NULL;
   return loc;
 }
 
@@ -68,13 +75,6 @@ default_emit_address (void *fail_arg __attribute__ ((unused)),
   obstack_printf (pool, AFORMAT, address);
 }
 
-static const char *
-dwarf_diename_integrate (Dwarf_Die *die)
-{
-  Dwarf_Attribute attr_mem;
-  return dwarf_formstring (dwarf_attr_integrate (die, DW_AT_name, &attr_mem));
-}
-
 /* Synthesize a new loc_address using the program on the obstack.  */
 static struct location *
 new_synthetic_loc (struct obstack *pool, struct location *origin, bool deref)
@@ -88,7 +88,7 @@ new_synthetic_loc (struct obstack *pool, struct location *origin, bool deref)
   loc->type = loc_address;
   loc->address.program = program;
   loc->address.stack_depth = 0;
-  loc->address.frame_base = NULL;
+  loc->address.declare = NULL;
   loc->address.used_deref = deref;
 
   if (origin->type == loc_register)
@@ -109,7 +109,7 @@ new_synthetic_loc (struct obstack *pool, struct location *origin, bool deref)
 /* Die in the middle of an expression.  */
 static struct location *
 lose (struct location *loc,
-      const char *failure, const Dwarf_Loc *lexpr, size_t i)
+      const char *failure, const Dwarf_Op *lexpr, size_t i)
 {
   FAIL (loc, N_("%s in DWARF expression [%Zu] at %" PRIu64
 		" (%#x: %" PRId64 ", %" PRId64 ")"),
@@ -131,7 +131,7 @@ lose (struct location *loc,
 
 static const char *
 translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
-	   const Dwarf_Loc *expr, const size_t len,
+	   const Dwarf_Op *expr, const size_t len,
 	   struct location *input,
 	   bool *need_fb, size_t *loser,
 	   struct location *loc)
@@ -192,6 +192,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 
   size_t i;
 
+  bool used_deref = false;
   inline const char *finish (struct location *piece)
     {
       if (stack_depth > 1)
@@ -201,9 +202,11 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	  obstack_1grow (pool, '\0');
 	  char *program = obstack_finish (pool);
 	  piece->type = loc_address;
+	  piece->address.declare = NULL;
 	  piece->address.program = program;
 	  piece->address.stack_depth = max_stack;
-	  piece->address.frame_base = NULL;
+	  piece->address.used_deref = used_deref;
+	  used_deref = false;
 	}
       else if (tos_register == -1)
 	DIE ("stack underflow");
@@ -219,6 +222,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 
   struct location *pieces = NULL, **tailpiece = &pieces;
   size_t piece_expr_start = 0;
+  Dwarf_Word piece_total_bytes = 0;
   for (i = 0; i < len; ++i)
     {
       unsigned int reg;
@@ -311,7 +315,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	  {
 	    POP (addr);
 	    push ("deref (sizeof (void *), " STACKFMT ")", addr);
-	    loc->address.used_deref = true;
+	    used_deref = true;
 	  }
 	  break;
 
@@ -320,7 +324,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	    POP (addr);
 	    push ("deref (" UFORMAT ", " STACKFMT ")",
 		  expr[i].number, addr);
-	    loc->address.used_deref = true;
+	    used_deref = true;
 	  }
 	  break;
 
@@ -330,7 +334,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	    POP (as);
 	    push ("xderef (sizeof (void *), " STACKFMT ", " STACKFMT ")",
 		  addr, as);
-	    loc->address.used_deref = true;
+	    used_deref = true;
 	  }
 	  break;
 
@@ -340,7 +344,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	    POP (as);
 	    push ("xderef (" UFORMAT ", " STACKFMT ", " STACKFMT ")",
 		  expr[i].number, addr, as);
-	    loc->address.used_deref = true;
+	    used_deref = true;
 	  }
 	  break;
 
@@ -469,16 +473,28 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 	    DIE ("DW_OP_piece left multiple values on stack");
 	  else
 	    {
-	      struct location *piece = alloc_location (pool, input);
-	      const char *failure = finish (piece);
+	      /* The obstack has a pending program for loc_address,
+		 so we must finish that piece off before we can
+		 allocate again.  */
+	      struct location temp_piece =
+		{
+		  .fail = loc->fail,
+		  .fail_arg = loc->fail_arg,
+		  .emit_address = loc->emit_address,
+		  .frame_base = NULL,
+		  .ops = &expr[piece_expr_start],
+		  .nops = i - piece_expr_start,
+		};
+	      const char *failure = finish (&temp_piece);
 	      if (failure != NULL)
 		return failure;
 
-	      piece->ops = &expr[piece_expr_start];
-	      piece->nops = i - piece_expr_start;
+	      struct location *piece = obstack_alloc (pool, sizeof *piece);
+	      *piece = temp_piece;
+
 	      piece_expr_start = i + 1;
 
-	      piece->byte_size = expr[i].number;
+	      piece_total_bytes += piece->byte_size = expr[i].number;
 
 	      *tailpiece = piece;
 	      tailpiece = &piece->next;
@@ -504,6 +520,7 @@ translate (struct obstack *pool, int indent, Dwarf_Addr addrbias,
 
   loc->type = loc_noncontiguous;
   loc->pieces = pieces;
+  loc->byte_size = piece_total_bytes;
 
   return NULL;
 
@@ -527,13 +544,15 @@ location_from_address (struct obstack *pool,
 		       void (*emit_address) (void *fail_arg,
 					     struct obstack *, Dwarf_Addr),
 		       int indent, Dwarf_Addr dwbias,
-		       const Dwarf_Loc *expr, size_t len, Dwarf_Addr address,
+		       const Dwarf_Op *expr, size_t len, Dwarf_Addr address,
 		       struct location **input, Dwarf_Attribute *fb_attr)
 {
   struct location *loc = obstack_alloc (pool, sizeof *loc);
   loc->fail = *input == NULL ? fail : (*input)->fail;
   loc->fail_arg = *input == NULL ? fail_arg : (*input)->fail_arg;
   loc->emit_address = *input == NULL ? emit_address : (*input)->emit_address;
+  loc->byte_size = 0;
+  loc->frame_base = NULL;
 
   bool need_fb = false;
   size_t loser;
@@ -548,10 +567,13 @@ location_from_address (struct obstack *pool,
       /* The main expression uses DW_OP_fbreg, so we need to compute
 	 the DW_AT_frame_base attribute expression's value first.  */
 
-      Dwarf_Loc *fb_expr;
+      if (fb_attr == NULL)
+	FAIL (loc, N_("required DW_AT_frame_base attribute not supplied"));
+
+      Dwarf_Op *fb_expr;
       size_t fb_len;
-      switch (dwarf_addrloclists (fb_attr, address - dwbias,
-				  &fb_expr, &fb_len, 1))
+      switch (dwarf_getlocation_addr (fb_attr, address - dwbias,
+				      &fb_expr, &fb_len, 1))
 	{
 	case 1:			/* Should always happen.  */
 	  if (fb_len == 0)
@@ -560,19 +582,19 @@ location_from_address (struct obstack *pool,
 
 	default:		/* Shouldn't happen.  */
 	case -1:
-	  FAIL (*input, N_("dwarf_addrloclists (form %#x): %s"),
+	  FAIL (loc, N_("dwarf_getlocation_addr (form %#x): %s"),
 		dwarf_whatform (fb_attr), dwarf_errmsg (-1));
 	  return NULL;
 
 	case 0:			/* Shouldn't happen.  */
 	fb_inaccessible:
-	  FAIL (*input, N_("DW_AT_frame_base not accessible at this address"));
+	  FAIL (loc, N_("DW_AT_frame_base not accessible at this address"));
 	  return NULL;
 	}
 
-      loc->address.frame_base = alloc_location (pool, loc);
+      loc->frame_base = alloc_location (pool, loc);
       failure = translate (pool, indent + 1, dwbias, fb_expr, fb_len, NULL,
-			   NULL, &loser, loc->address.frame_base);
+			   NULL, &loser, loc->frame_base);
       if (failure != NULL)
 	return lose (loc, failure, fb_expr, loser);
     }
@@ -592,7 +614,7 @@ location_from_address (struct obstack *pool,
 static struct location *
 location_relative (struct obstack *pool,
 		   int indent, Dwarf_Addr dwbias,
-		   const Dwarf_Loc *expr, size_t len, Dwarf_Addr address,
+		   const Dwarf_Op *expr, size_t len, Dwarf_Addr address,
 		   struct location **input, Dwarf_Attribute *fb_attr)
 {
   Dwarf_Sword *stack;
@@ -950,7 +972,7 @@ location_relative (struct obstack *pool,
    If DW_OP_fbreg is used, it may have a subfragment computing from
    the FB_ATTR location expression.
 
-   On errors, exit and never return (XXX ?).  On success, return the
+   On errors, call FAIL and never return.  On success, return the
    first fragment created, which is also chained onto (*INPUT)->next.
    *INPUT is then updated with the new tail of that chain.  */
 
@@ -961,34 +983,12 @@ c_translate_location (struct obstack *pool,
 		      void *fail_arg,
 		      void (*emit_address) (void *fail_arg,
 					    struct obstack *, Dwarf_Addr),
-		      int indent, Dwarf_Addr dwbias,
-		      Dwarf_Attribute *loc_attr, Dwarf_Addr address,
+		      int indent, Dwarf_Addr dwbias, Dwarf_Addr pc_address,
+		      const Dwarf_Op *expr, size_t len,
 		      struct location **input, Dwarf_Attribute *fb_attr)
 {
-  ++indent;
+  indent += 2;
 
-  Dwarf_Loc *expr;
-  size_t len;
-  switch (dwarf_addrloclists (loc_attr, address - dwbias, &expr, &len, 1))
-    {
-    case 1:			/* Should always happen.  */
-      if (len == 0)
-	goto inaccessible;
-      break;
-
-    default:			/* Shouldn't happen.  */
-    case -1:
-      FAIL (*input, N_("dwarf_addrloclists (form %#x): %s"),
-	    dwarf_whatform (fb_attr), dwarf_errmsg (-1));
-      return NULL;
-
-    case 0:			/* Shouldn't happen.  */
-    inaccessible:
-      FAIL (*input, N_("not accessible at this address"));
-      return NULL;
-    }
-
-  ++indent;
   switch (*input == NULL ? loc_address : (*input)->type)
     {
     case loc_address:
@@ -996,17 +996,16 @@ c_translate_location (struct obstack *pool,
 	 This expression will compute starting with that on the stack.  */
       return location_from_address (pool, fail, fail_arg,
 				    emit_address ?: &default_emit_address,
-				    indent, dwbias, expr, len, address,
+				    indent, dwbias, expr, len, pc_address,
 				    input, fb_attr);
 
     case loc_noncontiguous:
     case loc_register:
       /* The starting point is not an address computation, but a
 	 register.  We can only handle limited computations from here.  */
-      return location_relative (pool, indent, dwbias, expr, len, address,
+      return location_relative (pool, indent, dwbias, expr, len, pc_address,
 				input, fb_attr);
 
-    case loc_final:		/* Bogus caller.  */
     default:
       abort ();
       break;
@@ -1015,7 +1014,6 @@ c_translate_location (struct obstack *pool,
   return NULL;
 }
 
-
 /* Emit "uintNN_t TARGET = ...;".  */
 static bool
 emit_base_fetch (struct obstack *pool, Dwarf_Word byte_size,
@@ -1028,10 +1026,10 @@ emit_base_fetch (struct obstack *pool, Dwarf_Word byte_size,
     {
     case loc_address:
       if (byte_size != 0 && byte_size != (Dwarf_Word) -1)
-	obstack_printf (pool, "%s = deref (%" PRIu64 ", addr); ",
+	obstack_printf (pool, "%s = deref (%" PRIu64 ", addr);",
 			target, byte_size);
       else
-	obstack_printf (pool, "%s = deref (sizeof %s, addr); ",
+	obstack_printf (pool, "%s = deref (sizeof %s, addr);",
 			target, target);
       return true;
 
@@ -1040,8 +1038,7 @@ emit_base_fetch (struct obstack *pool, Dwarf_Word byte_size,
       break;
 
     case loc_noncontiguous:
-      /* Could be handled if it ever happened.  */
-      FAIL (loc, N_("noncontiguous locations not supported"));
+      FAIL (loc, N_("noncontiguous location for base fetch"));
       break;
 
     default:
@@ -1073,8 +1070,7 @@ emit_base_store (struct obstack *pool, Dwarf_Word byte_size,
       break;
 
     case loc_noncontiguous:
-      /* Could be handled if it ever happened.  */
-      FAIL (loc, N_("noncontiguous locations not supported"));
+      FAIL (loc, N_("noncontiguous location for base store"));
       break;
 
     default:
@@ -1085,42 +1081,130 @@ emit_base_store (struct obstack *pool, Dwarf_Word byte_size,
   return false;
 }
 
-/* Translate a fragment to dereference the given pointer type,
-   where *INPUT is the location of the pointer with that type.
 
-   We chain on a loc_address program that yields this pointer value
-   (i.e. the location of what it points to).  */
-
-void
-c_translate_pointer (struct obstack *pool, int indent,
-		     Dwarf_Addr dwbias __attribute__ ((unused)),
-		     Dwarf_Die *typedie, struct location **input)
+/* Slice up an object into pieces no larger than MAX_PIECE_BYTES,
+   yielding a loc_noncontiguous location unless LOC is small enough.  */
+static struct location *
+discontiguify (struct obstack *pool, int indent, struct location *loc,
+	       Dwarf_Word total_bytes, Dwarf_Word max_piece_bytes)
 {
-  assert (dwarf_tag (typedie) == DW_TAG_pointer_type);
-
-  ++indent;
-
-  obstack_printf (pool, "%*s{ ", (indent + 2) * 2, "");
-
-  bool deref = false;
-  Dwarf_Attribute attr_mem;
-  Dwarf_Word byte_size;
-  if (dwarf_attr_integrate (typedie, DW_AT_byte_size, &attr_mem) == NULL)
+  inline bool pieces_small_enough (void)
     {
-      obstack_printf (pool, "uintptr_t ");
-      deref = emit_base_fetch (pool, 0, "tmp", false, *input);
+      if (loc->type != loc_noncontiguous)
+	return (loc->byte_size ?: total_bytes) <= max_piece_bytes;
+      for (struct location *p = loc->pieces; p != NULL; p = p->next)
+	if (p->byte_size > max_piece_bytes)
+	  return false;
+      return true;
     }
-  else if (dwarf_formudata (&attr_mem, &byte_size) != 0)
-    FAIL (*input,
-	  N_("cannot get byte_size attribute for type %s: %s"),
-	  dwarf_diename_integrate (typedie) ?: "<anonymous>",
-	  dwarf_errmsg (-1));
-  else
-    deref = emit_base_fetch (pool, byte_size, "tmp", true, *input);
 
-  obstack_printf (pool, " addr = tmp; }\n");
+  if (pieces_small_enough ())
+    return loc;
 
-  struct location *loc = new_synthetic_loc (pool, *input, deref);
+  struct location *noncontig = alloc_location (pool, loc);
+  noncontig->next = NULL;
+  noncontig->type = loc_noncontiguous;
+  noncontig->byte_size = total_bytes;
+  noncontig->pieces = NULL;
+  struct location **tailpiece = &noncontig->pieces;
+  inline void add (struct location *piece)
+    {
+      *tailpiece = piece;
+      tailpiece = &piece->next;
+    }
+
+  switch (loc->type)
+    {
+    case loc_address:
+      {
+	/* Synthesize a piece that sets "container_addr" to the computed
+	   address of the whole object.  Each piece will refer to this.  */
+	obstack_printf (pool, "%*scontainer_addr = addr;\n",
+			indent++ * 2, "");
+	loc->next = new_synthetic_loc (pool, loc, false);
+	loc->next->byte_size = loc->byte_size;
+	loc->next->type = loc_fragment;
+	loc->next->address.declare = "container_addr";
+	loc = loc->next;
+
+	/* Synthesize pieces that just compute "container_addr + N".  */
+	Dwarf_Word offset = 0;
+	while (total_bytes - offset > 0)
+	  {
+	    Dwarf_Word size = total_bytes - offset;
+	    if (size > max_piece_bytes)
+	      size = max_piece_bytes;
+
+	    obstack_printf (pool, "%*saddr = container_addr + " UFORMAT ";\n",
+			    indent * 2, "", offset);
+	    struct location *piece = new_synthetic_loc (pool, loc, false);
+	    piece->byte_size = size;
+	    add (piece);
+
+	    offset += size;
+	  }
+
+	--indent;
+	break;
+      }
+
+    case loc_register:
+      FAIL (loc, N_("single register too big for fetch/store ???"));
+      break;
+
+    case loc_noncontiguous:
+      /* Could be handled if it ever happened.  */
+      FAIL (loc, N_("discontiguify of noncontiguous location not supported"));
+      break;
+
+    default:
+      abort ();
+      break;
+    }
+
+  loc->next = noncontig;
+  return noncontig;
+}
+
+/* Make a fragment that declares a union such as:
+    union {
+      char bytes[8];
+      struct {
+        uint32_t p0;
+        uint32_t p4;
+      } pieces __attribute__ ((packed));
+      uint64_t whole;
+    } u;
+*/
+static void
+declare_noncontig_union (struct obstack *pool, int indent,
+			 struct location **input, struct location *loc)
+{
+  obstack_printf (pool, "%*sunion {\n", indent++ * 2, "");
+
+  obstack_printf (pool, "%*schar bytes[%" PRIu64 "];\n",
+		  indent * 2, "", loc->byte_size);
+
+  obstack_printf (pool, "%*sstruct {\n", indent++ * 2, "");
+
+  Dwarf_Word offset = 0;
+  for (struct location *p = loc->pieces; p != NULL; p = p->next)
+    {
+      obstack_printf (pool, "%*suint%" PRIu64 "_t p%" PRIu64 ";\n",
+		      indent * 2, "", p->byte_size * 8, offset);
+      offset += p->byte_size;
+    }
+
+  obstack_printf (pool, "%*s} pieces __attribute__ ((packed));\n",
+		  --indent * 2, "");
+
+  obstack_printf (pool, "%*suint%" PRIu64 "_t whole;\n",
+		  indent * 2, "", loc->byte_size * 8);
+
+  obstack_printf (pool, "%*s} u;\n", --indent * 2, "");
+
+  loc = new_synthetic_loc (pool, *input, false);
+  loc->type = loc_decl;
   (*input)->next = loc;
   *input = loc;
 }
@@ -1140,80 +1224,62 @@ base_byte_size (Dwarf_Die *typedie, struct location *origin)
 
   FAIL (origin,
 	 N_("cannot get byte_size attribute for type %s: %s"),
-	 dwarf_diename_integrate (typedie) ?: "<anonymous>",
+	 dwarf_diename (typedie) ?: "<anonymous>",
 	 dwarf_errmsg (-1));
   return -1;
 }
 
-/* Emit a code fragment like:
-	{ uintNN_t tmp = ...; fetch_bitfield (TARGET, tmp, B0, Bn); }
-   or
-	{ uintNN_t tmp = ...; store_bitfield (tmp, TARGET, B0, Bn);
-	  ... = tmp; }
-*/
-static bool
-emit_bitfield (struct obstack *pool, int indent,
-	       Dwarf_Die *die, Dwarf_Word byte_size, struct location *loc,
-	       const char *target, bool store)
+/* Fetch the bitfield parameters.  */
+static void
+get_bitfield (struct location *loc,
+	      Dwarf_Die *die, Dwarf_Word *bit_offset, Dwarf_Word *bit_size)
 {
-  Dwarf_Word bit_offset, bit_size;
   Dwarf_Attribute attr_mem;
   if (dwarf_attr_integrate (die, DW_AT_bit_offset, &attr_mem) == NULL
-      || dwarf_formudata (&attr_mem, &bit_offset) != 0
+      || dwarf_formudata (&attr_mem, bit_offset) != 0
       || dwarf_attr_integrate (die, DW_AT_bit_size, &attr_mem) == NULL
-      || dwarf_formudata (&attr_mem, &bit_size) != 0)
+      || dwarf_formudata (&attr_mem, bit_size) != 0)
     FAIL (loc, N_("cannot get bit field parameters: %s"), dwarf_errmsg (-1));
-
-  /* Emit "{ uintNN_t tmp = ...;" to fetch the base type.  */
-
-  obstack_printf (pool, "%*s{ ", indent * 2, "");
-  bool deref = emit_base_fetch (pool, byte_size, "tmp", true, loc);
-
-  if (store)
-    {
-      obstack_printf (pool,
-		      "\n%*s"
-		      "store_bitfield (tmp, %s, %" PRIu64 ", %" PRIu64 ");\n"
-		      "\n%*s",
-		      indent * 2, "",
-		      target, bit_offset, bit_size,
-		      (indent + 1) * 2, "");
-      deref = emit_base_store (pool, byte_size, "tmp", loc) || deref;
-      obstack_printf (pool, "}\n");
-    }
-  else
-    obstack_printf (pool,
-		    "fetch_bitfield (%s, tmp, %" PRIu64 ", %" PRIu64 "); }\n",
-		    target, bit_offset, bit_size);
-
-  return deref;
 }
 
-/* Translate a fragment to fetch the value of variable or member DIE
+/* Translate a fragment to fetch the base-type value of BYTE_SIZE bytes
    at the *INPUT location and store it in lvalue TARGET.  */
-
-void
-c_translate_fetch (struct obstack *pool, int indent,
-		   Dwarf_Addr dwbias __attribute__ ((unused)),
-		   Dwarf_Die *die, Dwarf_Die *typedie,
-		   struct location **input, const char *target)
+static void
+translate_base_fetch (struct obstack *pool, int indent, Dwarf_Word byte_size,
+		      struct location **input, const char *target)
 {
-  ++indent;
-
-  Dwarf_Attribute size_attr;
-  Dwarf_Word byte_size;
-  if (dwarf_attr_integrate (die, DW_AT_byte_size, &size_attr) == NULL
-      || dwarf_formudata (&size_attr, &byte_size) != 0)
-    byte_size = base_byte_size (typedie, *input);
-
   bool deref = false;
-  if (dwarf_hasattr_integrate (die, DW_AT_bit_offset))
-    /* This is a bit field.  */
-    deref = emit_bitfield (pool, indent, die, byte_size, *input,
-			   target, false);
+
+  if ((*input)->type == loc_noncontiguous)
+    {
+      struct location *p = (*input)->pieces;
+
+      declare_noncontig_union (pool, indent, input, *input);
+
+      Dwarf_Word offset = 0;
+      char piece[sizeof "u.pieces.p" + 20] = "u.pieces.p";
+      while (p != NULL)
+	{
+	  struct location *newp = obstack_alloc (pool, sizeof *newp);
+	  *newp = *p;
+	  newp->next = NULL;
+	  (*input)->next = newp;
+	  *input = newp;
+
+	  snprintf (&piece[sizeof "u.pieces.p" - 1], 20, "%" PRIu64, offset);
+	  translate_base_fetch (pool, indent, p->byte_size, input, piece);
+	  (*input)->type = loc_fragment;
+
+	  offset += p->byte_size;
+	  p = p->next;
+	}
+
+      obstack_printf (pool, "%*s%s = u.whole;\n", indent * 2, "", target);
+    }
   else
     switch (byte_size)
       {
+      case 0:			/* Special case, means address size.  */
       case 1:
       case 2:
       case 4:
@@ -1230,11 +1296,135 @@ c_translate_fetch (struct obstack *pool, int indent,
       }
 
   struct location *loc = new_synthetic_loc (pool, *input, deref);
+  loc->byte_size = byte_size;
   loc->type = loc_final;
   (*input)->next = loc;
   *input = loc;
 }
 
+/* Determine the maximum size of a base type, from some DIE in the CU.  */
+static Dwarf_Word
+max_fetch_size (struct location *loc, Dwarf_Die *die)
+{
+  Dwarf_Die cu_mem;
+  uint8_t address_size;
+  Dwarf_Die *cu = dwarf_diecu (die, &cu_mem, &address_size, NULL);
+  if (cu == NULL)
+    FAIL (loc, N_("cannot determine CU address size from %s: %s"),
+	  dwarf_diename (die), dwarf_errmsg (-1));
+
+  return address_size;
+}
+
+/* Translate a fragment to fetch the value of variable or member DIE
+   at the *INPUT location and store it in lvalue TARGET.  */
+void
+c_translate_fetch (struct obstack *pool, int indent,
+		   Dwarf_Addr dwbias __attribute__ ((unused)),
+		   Dwarf_Die *die, Dwarf_Die *typedie,
+		   struct location **input, const char *target)
+{
+  ++indent;
+
+  Dwarf_Attribute size_attr;
+  Dwarf_Word byte_size;
+  if (dwarf_attr_integrate (die, DW_AT_byte_size, &size_attr) == NULL
+      || dwarf_formudata (&size_attr, &byte_size) != 0)
+    byte_size = base_byte_size (typedie, *input);
+
+  *input = discontiguify (pool, indent, *input, byte_size,
+			  max_fetch_size (*input, die));
+
+  if (dwarf_hasattr_integrate (die, DW_AT_bit_offset))
+    {
+      /* This is a bit field.  Fetch the containing base type into a
+	 temporary variable.  */
+
+      translate_base_fetch (pool, indent, byte_size, input, "tmp");
+      (*input)->type = loc_fragment;
+      (*input)->address.declare = "tmp";
+
+      Dwarf_Word bit_offset, bit_size;
+      get_bitfield (*input, die, &bit_offset, &bit_size);
+
+      obstack_printf (pool, "%*s"
+		      "fetch_bitfield (%s, tmp, %" PRIu64 ", %" PRIu64 ");\n",
+		      indent *2, "", target, bit_offset, bit_size);
+
+      struct location *loc = new_synthetic_loc (pool, *input, false);
+      loc->type = loc_final;
+      (*input)->next = loc;
+      *input = loc;
+    }
+  else
+    translate_base_fetch (pool, indent, byte_size, input, target);
+}
+
+/* Translate a fragment to store RVALUE into the base-type value of
+   BYTE_SIZE bytes at the *INPUT location.  */
+static void
+translate_base_store (struct obstack *pool, int indent, Dwarf_Word byte_size,
+		      struct location **input, struct location *store_loc,
+		      const char *rvalue)
+{
+  bool deref = false;
+
+  if (store_loc->type == loc_noncontiguous)
+    {
+      declare_noncontig_union (pool, indent, input, store_loc);
+
+      obstack_printf (pool, "%*su.whole = %s;\n", indent * 2, "", rvalue);
+      struct location *loc = new_synthetic_loc (pool, *input, deref);
+      loc->type = loc_fragment;
+      (*input)->next = loc;
+      *input = loc;
+
+      Dwarf_Word offset = 0;
+      char piece[sizeof "u.pieces.p" + 20] = "u.pieces.p";
+      for (struct location *p = store_loc->pieces; p != NULL; p = p->next)
+	{
+	  struct location *newp = obstack_alloc (pool, sizeof *newp);
+	  *newp = *p;
+	  newp->next = NULL;
+	  (*input)->next = newp;
+	  *input = newp;
+
+	  snprintf (&piece[sizeof "u.pieces.p" - 1], 20, "%" PRIu64, offset);
+	  translate_base_store (pool, indent,
+				p->byte_size, input, *input, piece);
+	  (*input)->type = loc_fragment;
+
+	  offset += p->byte_size;
+	}
+
+      (*input)->type = loc_final;
+    }
+  else
+    {
+      switch (byte_size)
+	{
+	case 1:
+	case 2:
+	case 4:
+	case 8:
+	  obstack_printf (pool, "%*s", indent * 2, "");
+	  deref = emit_base_store (pool, byte_size, rvalue, store_loc);
+	  obstack_printf (pool, "\n");
+	  break;
+
+	default:
+	  /* Could handle this generating call to memcpy equivalent.  */
+	  FAIL (*input, N_("store is larger than base integer types"));
+	  break;
+	}
+
+      struct location *loc = new_synthetic_loc (pool, *input, deref);
+      loc->type = loc_final;
+      (*input)->next = loc;
+      *input = loc;
+    }
+}
+
 /* Translate a fragment to fetch the value of variable or member DIE
    at the *INPUT location and store it in rvalue RVALUE.  */
 
@@ -1252,32 +1442,65 @@ c_translate_store (struct obstack *pool, int indent,
       || dwarf_formudata (&size_attr, &byte_size) != 0)
     byte_size = base_byte_size (typedie, *input);
 
-  bool deref = false;
+  *input = discontiguify (pool, indent, *input, byte_size,
+			  max_fetch_size (*input, die));
+
+  struct location *store_loc = *input;
+
   if (dwarf_hasattr_integrate (die, DW_AT_bit_offset))
-    /* This is a bit field.  */
-    deref = emit_bitfield (pool, indent, die, byte_size, *input, rvalue, true);
-  else
-    switch (byte_size)
-      {
-      case 1:
-      case 2:
-      case 4:
-      case 8:
-	obstack_printf (pool, "%*s", indent * 2, "");
-	deref = emit_base_store (pool, byte_size, rvalue, *input);
-	obstack_printf (pool, "\n");
-	break;
+    {
+      /* This is a bit field.  Fetch the containing base type into a
+	 temporary variable.  */
 
-      default:
-	/* Could handle this generating call to memcpy equivalent.  */
-	FAIL (*input, N_("fetch is larger than base integer types"));
-	break;
-      }
+      translate_base_fetch (pool, indent, byte_size, input, "tmp");
+      (*input)->type = loc_fragment;
+      (*input)->address.declare = "tmp";
 
-  struct location *loc = new_synthetic_loc (pool, *input, deref);
-  loc->type = loc_final;
-  (*input)->next = loc;
-  *input = loc;
+      Dwarf_Word bit_offset, bit_size;
+      get_bitfield (*input, die, &bit_offset, &bit_size);
+
+      obstack_printf (pool, "%*s"
+		      "store_bitfield (tmp, %s, %" PRIu64 ", %" PRIu64 ");\n",
+		      indent * 2, "", rvalue, bit_offset, bit_size);
+
+      struct location *loc = new_synthetic_loc (pool, *input, false);
+      loc->type = loc_fragment;
+      (*input)->next = loc;
+      *input = loc;
+
+      /* We have mixed RVALUE into the bits in "tmp".
+	 Now we'll store "tmp" back whence we fetched it.  */
+      rvalue = "tmp";
+    }
+
+  translate_base_store (pool, indent, byte_size, input, store_loc, rvalue);
+}
+
+/* Translate a fragment to dereference the given pointer type,
+   where *INPUT is the location of the pointer with that type.
+
+   We chain on a loc_address program that yields this pointer value
+   (i.e. the location of what it points to).  */
+
+void
+c_translate_pointer (struct obstack *pool, int indent,
+		     Dwarf_Addr dwbias __attribute__ ((unused)),
+		     Dwarf_Die *typedie, struct location **input)
+{
+  assert (dwarf_tag (typedie) == DW_TAG_pointer_type);
+
+  Dwarf_Attribute attr_mem;
+  Dwarf_Word byte_size;
+  if (dwarf_attr_integrate (typedie, DW_AT_byte_size, &attr_mem) == NULL)
+    byte_size = 0;
+  else if (dwarf_formudata (&attr_mem, &byte_size) != 0)
+    FAIL (*input,
+	  N_("cannot get byte_size attribute for type %s: %s"),
+	  dwarf_diename (typedie) ?: "<anonymous>",
+	  dwarf_errmsg (-1));
+
+  translate_base_fetch (pool, indent + 1, byte_size, input, "addr");
+  (*input)->type = loc_address;
 }
 
 
@@ -1326,7 +1549,7 @@ array_stride (Dwarf_Die *typedie, struct location *origin)
       if (dwarf_formudata (&attr_mem, &stride) == 0)
 	return stride;
       FAIL (origin, N_("cannot get stride_size attribute array type %s: %s"),
-	    dwarf_diename_integrate (typedie) ?: "<anonymous>",
+	    dwarf_diename (typedie) ?: "<anonymous>",
 	    dwarf_errmsg (-1));
     }
 
@@ -1334,7 +1557,7 @@ array_stride (Dwarf_Die *typedie, struct location *origin)
   if (dwarf_attr_integrate (typedie, DW_AT_type, &attr_mem) == NULL
       || dwarf_formref_die (&attr_mem, &die_mem) == NULL)
     FAIL (origin, N_("cannot get element type of array type %s: %s"),
-	  dwarf_diename_integrate (typedie) ?: "<anonymous>",
+	  dwarf_diename (typedie) ?: "<anonymous>",
 	  dwarf_errmsg (-1));
 
   if (dwarf_attr_integrate (&die_mem, DW_AT_byte_size, &attr_mem) != NULL)
@@ -1344,7 +1567,7 @@ array_stride (Dwarf_Die *typedie, struct location *origin)
 	return stride;
       FAIL (origin,
 	    N_("cannot get byte_size attribute for array element type %s: %s"),
-	    dwarf_diename_integrate (&die_mem) ?: "<anonymous>",
+	    dwarf_diename (&die_mem) ?: "<anonymous>",
 	    dwarf_errmsg (-1));
     }
 
@@ -1501,9 +1724,6 @@ emit_loc_value (FILE *out, struct location *loc, unsigned int indent,
       break;
 
     case loc_address:
-      if (loc->address.frame_base != NULL)
-	emit_loc_value (out, loc->address.frame_base, indent,
-			"frame_base", true);
       emit_loc_address (out, loc, indent, target);
       break;
     }
@@ -1516,17 +1736,55 @@ c_emit_location (FILE *out, struct location *loc, int indent)
 {
   emit ("%*s{\n", indent * 2, "");
 
+  bool declared_addr = false;
+  for (struct location *l = loc; l != NULL; l = l->next)
+    switch (l->type)
+      {
+      case loc_decl:
+	emit ("%s", l->address.program);
+	break;
+
+      case loc_address:
+	if (declared_addr)
+	  break;
+	declared_addr = true;
+	l->address.declare = "addr";
+      case loc_fragment:
+      case loc_final:
+	if (l->address.declare != NULL)
+	  {
+	    if (l->byte_size == 0 || l->byte_size == (Dwarf_Word) -1)
+	      emit ("%*s%s %s;\n", (indent + 1) * 2, "",
+		    STACK_TYPE, l->address.declare);
+	    else
+	      emit ("%*suint%" PRIu64 "_t %s;\n", (indent + 1) * 2, "",
+		    l->byte_size * 8, l->address.declare);
+	  }
+
+      default:
+	break;
+      }
+
   bool deref = false;
-  for (bool declare_addr = true; loc->next != NULL; loc = loc->next)
+
+  if (loc->frame_base != NULL)
+    emit_loc_value (out, loc->frame_base, indent, "frame_base", true);
+
+  for (; loc->next != NULL; loc = loc->next)
     switch (loc->type)
       {
       case loc_address:
 	/* Emit the program fragment to calculate the address.  */
-	emit_loc_value (out, loc, indent + 1, "addr", declare_addr);
-	declare_addr = false;
+	emit_loc_value (out, loc, indent + 1, "addr", false);
 	deref = deref || loc->address.used_deref;
 	break;
 
+      case loc_fragment:
+	emit ("%s", loc->address.program);
+	deref = deref || loc->address.used_deref;
+	break;
+
+      case loc_decl:
       case loc_register:
       case loc_noncontiguous:
 	/* These don't produce any code directly.
@@ -1544,7 +1802,7 @@ c_emit_location (FILE *out, struct location *loc, int indent)
 
   emit ("%s%*s}\n", loc->address.program, indent * 2, "");
 
-  return deref;
+  return deref || loc->address.used_deref;
 }
 
 #undef emit
