@@ -46,7 +46,10 @@ std::ostream& operator << (std::ostream& o, const token& t);
 
 struct parse_error: public std::runtime_error
 {
-  parse_error (const std::string& msg): runtime_error (msg) {}
+  const token* tok;
+  parse_error (const std::string& msg): runtime_error (msg), tok (0) {}
+  parse_error (const std::string& msg, const token* t): runtime_error (msg),
+                                                        tok (t) {}
 };
 
 
@@ -70,21 +73,26 @@ private:
 class parser
 {
 public:
-  parser (std::istream& i, bool p);
-  parser (const std::string& n, bool p);
+  parser (systemtap_session& s, std::istream& i, bool p);
+  parser (systemtap_session& s, const std::string& n, bool p);
   ~parser ();
 
   stapfile* parse ();
 
-  static stapfile* parse (std::istream& i, bool privileged);
-  static stapfile* parse (const std::string& n, bool privileged);
+  static stapfile* parse (systemtap_session& s, std::istream& i, bool privileged);
+  static stapfile* parse (systemtap_session& s, const std::string& n, bool privileged);
 
 private:
+  systemtap_session& session;
   std::string input_name;
   std::istream* free_input;
   lexer input;
   bool privileged;
 
+  // preprocessing subordinate
+  std::vector<const token*> enqueued_pp;
+  const token* scan_pp ();
+
   // scanning state
   const token* last ();
   const token* next ();
@@ -100,6 +108,7 @@ private:
   // convenience forms
   const token* expect_op (std::string const & expected);
   const token* expect_kw (std::string const & expected);
+  const token* expect_number (int64_t & expected);
   const token* expect_ident (std::string & target);
   bool peek_op (std::string const & op);
   bool peek_kw (std::string const & kw);
@@ -126,6 +135,8 @@ private: // nonterminals
   next_statement* parse_next_statement ();
   break_statement* parse_break_statement ();
   continue_statement* parse_continue_statement ();
+  indexable* parse_indexable ();
+  const token *parse_hist_op_or_bare_name (hist_op *&hop, std::string &name);
   expression* parse_expression ();
   expression* parse_assignment ();
   expression* parse_ternary ();
