@@ -1,4 +1,13 @@
-#ifndef _STRING_C_ /* -*- linux-c -*- */
+/* -*- linux-c -*- 
+ * String Functions
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+#ifndef _STRING_C_
 #define _STRING_C_
 
 #include <linux/config.h>
@@ -62,6 +71,7 @@ void _stp_sprintf (String str, const char *fmt, ...)
 				_stp_pbuf_len[cpu] = STP_PRINT_BUF_LEN;
 				_stp_print_flush();
 			} else {
+				*buf ='\0';
 				_stp_print_flush();
 				va_start(args, fmt);
 				_stp_vsprintf(_stp_stdout, fmt, args); 
