@@ -1,4 +1,12 @@
-#ifndef _STRING_H_ /* -*- linux-c -*- */
+/* -*- linux-c -*-
+ * Copyright (C) 2005 Red Hat Inc.
+ *
+ * This file is part of systemtap, and is free software.  You can
+ * redistribute it and/or modify it under the terms of the GNU General
+ * Public License (GPL); either version 2, or (at your option) any
+ * later version.
+ */
+#ifndef _STRING_H_
 #define _STRING_H_
 
 /** Maximum string size allowed in Strings */
