@@ -1,5 +1,5 @@
 // tapset resolution
-// Copyright (C) 2005 Red Hat Inc.
+// Copyright (C) 2005, 2006 Red Hat Inc.
 // Copyright (C) 2005 Intel Corporation.
 //
 // This file is part of systemtap, and is free software.  You can
@@ -21,6 +21,7 @@
 #include <stdexcept>
 #include <vector>
 #include <cstdarg>
+#include <cassert>
 
 extern "C" {
 #include <fcntl.h>
@@ -84,6 +85,63 @@ lex_cast_qstring(IN const & in)
 }
 
 
+// ------------------------------------------------------------------------
+
+void
+derived_probe::emit_probe_prologue (translator_output* o,
+                                    const std::string& statereq)
+{
+  o->newline() << "struct context* c = per_cpu_ptr "
+               << "(contexts, smp_processor_id());";
+  o->newline() << "if (atomic_read (&session_state) != " << statereq << ")";
+  o->newline(1) << "goto probe_epilogue;";
+
+  o->newline(-1) << "if (unlikely (atomic_inc_return (&c->busy) != 1)) {";
+
+  o->newline(1) << "if (atomic_inc_return (& skipped_count) > MAXSKIPPED) {";
+  o->newline(1) << "atomic_set (& session_state, STAP_SESSION_ERROR);";
+  // NB: We don't assume that we can safely call stp_error etc. in such
+  // a reentrant context.  But this is OK:
+  o->newline() << "_stp_exit ();";
+  o->newline(-1) << "}";
+
+  o->newline() << "atomic_dec (& c->busy);";
+  o->newline() << "goto probe_epilogue;";
+  o->newline(-1) << "}";
+  o->newline();
+  o->newline() << "c->last_error = 0;";
+  o->newline() << "c->probe_point = probe_point;";
+  o->newline() << "c->nesting = 0;";
+  o->newline() << "c->regs = 0;";
+  o->newline() << "c->actioncount = 0;";
+}
+
+void
+derived_probe::emit_probe_epilogue (translator_output* o)
+{
+  o->newline() << "if (unlikely (c->last_error && c->last_error[0])) {";
+  o->newline(1) << "if (c->last_stmt != NULL)";
+  o->newline(1) << "_stp_softerror (\"%s near %s\", c->last_error, c->last_stmt);";
+  o->newline(-1) << "else";
+  o->newline(1) << "_stp_softerror (\"%s\", c->last_error);";
+  o->indent(-1);
+  o->newline() << "atomic_inc (& error_count);";
+
+  o->newline() << "if (atomic_read (& error_count) > MAXERRORS) {";
+  o->newline(1) << "atomic_set (& session_state, STAP_SESSION_ERROR);";
+  o->newline() << "_stp_exit ();";
+  o->newline(-1) << "}";
+
+  o->newline(-1) << "}";
+  
+  o->newline() << "atomic_dec (&c->busy);";
+  o->newline(-1) << "probe_epilogue: ;";
+  o->indent(1);
+}
+
+
+
+
 // ------------------------------------------------------------------------
 // begin/end probes are run right during registration / deregistration
 // ------------------------------------------------------------------------
@@ -146,41 +204,20 @@ be_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
 
       // While begin/end probes are executed single-threaded, we
       // still code defensively and use a per-cpu context.
-      o->newline(1) << "struct context* c = per_cpu_ptr (contexts, smp_processor_id());";
+      o->indent(1);
       o->newline() << "const char* probe_point = "
-		   << lex_cast_qstring(*l) << ";";
-
-      // A precondition for running a probe handler is that we're in STARTING
-      // or STOPPING state (not ERROR), and that no one else is already using
-      // this context.
-      o->newline() << "if (atomic_read (&session_state) != ";
-      if (begin) o->line() << "STAP_SESSION_STARTING)";
-      else o->line() << "STAP_SESSION_STOPPING)";
-      o->newline(1) << "return;";
-      o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
-      o->newline(1) << "_stp_warn (\"probe reentrancy (%s vs %s)\\n\", "
-		    << "c->probe_point, probe_point);";
-      o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-      o->newline() << "atomic_dec (&c->busy);";
-      o->newline() << "return;";
-      o->newline(-1) << "}";
-      o->newline();
-      o->newline() << "c->last_error = 0;";
-      o->newline() << "c->probe_point = probe_point;";
-      o->newline() << "c->nesting = 0;";
-      o->newline() << "c->regs = 0;";
-      o->newline() << "c->actioncount = 0;";
+                   << lex_cast_qstring(*l) << ";";
+      emit_probe_prologue (o,
+                           (begin ?
+                            "STAP_SESSION_STARTING" :
+                            "STAP_SESSION_STOPPING"));
 
       // NB: locals are initialized by probe function itself
       o->newline() << "probe_" << j << " (c);";
 
-      o->newline() << "if (c->last_error && c->last_error[0]) {";
-      o->newline(1) << "_stp_error (\"%s near %s\", c->last_error, c->last_stmt);";
-      o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-      o->newline(-1) << "}";
+      emit_probe_epilogue (o);
 
-      o->newline() << "atomic_dec (&c->busy);";
-      o->newline(-1) << "}" << endl;
+      o->newline(-1) << "}\n";
     }
 }
 
@@ -234,6 +271,143 @@ inline_instance_info
   Dwarf_Die die;
 };
 
+class
+symbol_cache
+{
+  // For each module, we keep a multimap from function names to
+  // (cudie, funcdie*) pairs.  The first time we pass over a module,
+  // we build up this multimap as an index. Our iteration over the
+  // module's CUs and functions is then driven by the function or
+  // statement pattern string we're scanning for.
+  struct entry
+  {
+    Dwarf_Die cu;
+    Dwarf_Die function;
+  };
+  typedef multimap<string, entry> index;
+  map<Dwarf *, index*> indices;
+  index *curr_index;
+  Dwarf_Die * cu_die;
+  void make_entry_for_function(Dwarf_Die *func_die);
+  static int function_callback(Dwarf_Die * func, void * arg);
+  void index_module(Dwarf * mod);
+public:  
+  void select_die_subsets(Dwarf * mod,
+			  string const & pattern,
+			  set<Dwarf_Die> & cus,
+			  multimap<Dwarf_Die, Dwarf_Die> & funcs);
+};
+
+void
+symbol_cache::make_entry_for_function(Dwarf_Die *func_die)
+{
+  entry e;
+  assert(this->cu_die);
+  assert(this->curr_index);
+  e.cu = *(this->cu_die);
+  e.function = *(func_die);
+  char const * fname = dwarf_diename(func_die);
+  if (fname)
+    curr_index->insert(make_pair(string(fname), e));
+}
+
+int
+symbol_cache::function_callback(Dwarf_Die * func, void * arg)
+{
+  symbol_cache *sym = static_cast<symbol_cache*>(arg);
+  sym->make_entry_for_function(func);
+  return DWARF_CB_OK;
+}
+
+void
+symbol_cache::index_module(Dwarf *module_dwarf)
+{
+  Dwarf_Off off = 0;
+  size_t cuhl = 0;
+  Dwarf_Off noff = 0;
+  this->cu_die = NULL;
+  while (dwarf_nextcu (module_dwarf, off, &noff, &cuhl, NULL, NULL, NULL) == 0)
+    {
+      Dwarf_Die die_mem;
+      this->cu_die = dwarf_offdie (module_dwarf, off + cuhl, &die_mem);
+      dwarf_getfuncs (this->cu_die, function_callback, this, 0);
+      off = noff;
+    }
+  this->cu_die = NULL;
+}
+
+inline bool
+operator<(Dwarf_Die const & a, 
+	  Dwarf_Die const & b)
+{
+  return (a.addr < b.addr)
+    || ((a.addr == b.addr) && (a.cu < b.cu))
+    || ((a.addr == b.addr) && (a.cu == b.cu) && (a.abbrev < b.abbrev));
+}
+
+inline bool
+operator==(Dwarf_Die const & a, 
+	   Dwarf_Die const & b)
+{
+  return !((a < b) || (b < a));
+}
+
+void 
+symbol_cache::select_die_subsets(Dwarf *mod,
+				 string const & pattern,
+				 set<Dwarf_Die> & cus,
+				 multimap<Dwarf_Die, Dwarf_Die> & funcs)
+{
+  cus.clear();
+  funcs.clear();
+  index *ix = NULL;
+
+  // First find the index for this module. If there's no index, build
+  // one.
+  map<Dwarf *, index*>::const_iterator i = indices.find(mod);
+  if (i == indices.end())
+    {
+      this->curr_index = new index;
+      index_module(mod);      
+      indices.insert(make_pair(mod, this->curr_index));
+      ix = this->curr_index;
+      this->curr_index = NULL;
+      this->cu_die = NULL;
+    }
+  else
+    ix = i->second;
+
+  assert(ix);
+
+  // Now stem the pattern such that we have a minimal non-wildcard
+  // prefix to search in the multimap for. We will use the full pattern
+  // to narrow this set further.
+  string stem;
+  for (string::const_iterator i = pattern.begin();
+       i != pattern.end(); ++i)
+    {
+      if (*i == '?' || *i == '*' || *i == '[' || *i == ']')
+	break;
+      stem += *i;
+    }
+
+  // Now perform a lower-bound on the multimap, refine that result
+  // set, and copy the CU and function DIEs into the parameter sets.
+  index::const_iterator j = stem.empty() ? ix->begin() : ix->lower_bound(stem);
+  while (j != ix->end() && 
+	 (stem.empty() || j->first.compare(0, stem.size(), stem) == 0))
+    {
+      if (fnmatch(pattern.c_str(), j->first.c_str(), 0) == 0)
+	{
+	  cus.insert(j->second.cu);
+	  funcs.insert(make_pair(j->second.cu, j->second.function));
+	}
+      ++j;	
+    }
+}
+			   
+
+
 static int
 query_cu (Dwarf_Die * cudie, void * arg);
 
@@ -254,6 +428,8 @@ dwflpp
   systemtap_session & sess;
   Dwfl * dwfl;
 
+  symbol_cache cache;
+
   // These are "current" values we focus on.
   Dwfl_Module * module;
   Dwarf * module_dwarf;
@@ -266,6 +442,9 @@ dwflpp
   Dwarf_Die * cu;
   Dwarf_Die * function;
 
+  set<Dwarf_Die> pattern_limited_cus;
+  multimap<Dwarf_Die, Dwarf_Die> pattern_limited_funcs;
+
   string module_name;
   string cu_name;
   string function_name;
@@ -277,7 +456,7 @@ dwflpp
     if (in)
       return in;
     if (false && sess.verbose)
-      clog << "WARNING: no name found for " << type << endl;
+      clog << "WARNING: no name found for " << type << "\n";
     return string("");
   }
 
@@ -303,10 +482,17 @@ dwflpp
 	if (required)
 	  throw semantic_error (msg);
 	else
-	  cerr << "WARNING: " << msg << endl;
+	  cerr << "WARNING: " << msg << "\n";
       }
   }
 
+  void limit_search_to_function_pattern(string const & pattern)
+  {
+    get_module_dwarf(true);
+    cache.select_die_subsets(module_dwarf, pattern, 
+			     pattern_limited_cus, 
+			     pattern_limited_funcs);
+  }
 
   void focus_on_module(Dwfl_Module * m)
   {
@@ -322,6 +508,9 @@ dwflpp
 
     module_dwarf = NULL;
 
+    pattern_limited_cus.clear();
+    pattern_limited_funcs.clear();
+    
     cu_name.clear();
     cu = NULL;
 
@@ -360,8 +549,6 @@ dwflpp
   {
     assert(dwfl);
     cu = NULL;
-    if (false && sess.verbose)
-      clog << "focusing on module containing global addr " << a << endl;
     Dwfl_Module* mod = dwfl_addrmodule(dwfl, a);
     if (mod) // address could be wildly out of range
       focus_on_module(mod);
@@ -397,7 +584,7 @@ dwflpp
     if (false && sess.verbose)
       clog << "module addr " << hex << a
 	   << " + module start " << module_start
-	   << " -> global addr " << (a + module_start) << dec << endl;
+	   << " -> global addr " << (a + module_start) << dec << "\n";
     return a + module_start;
   }
 
@@ -409,7 +596,7 @@ dwflpp
     if (false && sess.verbose)
       clog << "global addr " << a
 	   << " - module start " << hex << module_start
-	   << " -> module addr " << (a - module_start) << dec << endl;
+	   << " -> module addr " << (a - module_start) << dec << "\n";
     return a - module_bias;
   }
 
@@ -421,7 +608,7 @@ dwflpp
     if (t && sess.verbose)
       clog << "pattern '" << pattern << "' "
 	   << "matches "
-	   << "module '" << module_name << "'" << endl;
+	   << "module '" << module_name << "'" << "\n";
     return t;
   }
 
@@ -433,7 +620,7 @@ dwflpp
     if (t && sess.verbose)
       clog << "pattern '" << pattern << "' "
 	   << "matches "
-	   << "function '" << function_name << "'" << endl;
+	   << "function '" << function_name << "'" << "\n";
     return t;
   }
 
@@ -445,7 +632,7 @@ dwflpp
     if (t && sess.verbose)
       clog << "pattern '" << pattern << "' "
 	   << "matches "
-	   << "CU '" << cu_name << "'" << endl;
+	   << "CU '" << cu_name << "'" << "\n";
     return t;
   }
 
@@ -555,18 +742,12 @@ dwflpp
     if (!module_dwarf)
       return;
 
-    Dwarf *dw = module_dwarf;
-    Dwarf_Off off = 0;
-    size_t cuhl;
-    Dwarf_Off noff;
-    while (dwarf_nextcu (dw, off, &noff, &cuhl, NULL, NULL, NULL) == 0)
+    for (set<Dwarf_Die>::const_iterator i = pattern_limited_cus.begin();
+	 i != pattern_limited_cus.end(); ++i)
       {
-	Dwarf_Die die_mem;
-	Dwarf_Die *die;
-	die = dwarf_offdie (dw, off + cuhl, &die_mem);
-	if (callback (die, data) != DWARF_CB_OK)
+	Dwarf_Die die = *i;
+	if (callback (&die, data) != DWARF_CB_OK)
 	  break;
-	off = noff;
       }
   }
 
@@ -592,7 +773,15 @@ dwflpp
   {
     assert (module);
     assert (cu);
-    dwarf_getfuncs (cu, callback, data, 0);
+    multimap<Dwarf_Die, Dwarf_Die>::const_iterator i = pattern_limited_funcs.lower_bound(*cu);
+    while (i != pattern_limited_funcs.end() && (i->first == *cu))
+      {
+	Dwarf_Die func_die = i->second;
+	if (callback (&func_die, data) != DWARF_CB_OK)
+	  break;
+
+	++i;
+      }
   }
 
 
@@ -702,7 +891,7 @@ dwflpp
 	  {
 	    filtered_srcfiles.insert (fname);
 	    if (sess.verbose)
-	      clog << "selected source file '" << fname << "'" << endl;
+	      clog << "selected source file '" << fname << "'\n";
 	  }
       }
     }
@@ -750,12 +939,22 @@ dwflpp
     // This heuristic attempts to pick the first address that has a
     // source line distinct from the function declaration's (entrypc's).
     // This should be the first statement *past* the prologue.
+    //
+    // However, for some tail-call optimized functions
+    // ("sys_exit_group" in linux 2.6), there might be only a single
+    // line record for the function.  We guess at this situation if
+    // the "next" line record refers to an address past the
+    // dwarf_highpc of the current function.  In this case, we back
+    // down to the address from the previous line record.
+    //
     assert(module);
     assert(cu);
 
-    size_t nlines;
-    Dwarf_Lines *lines;
-    Dwarf_Addr last_function_entrypc;
+    size_t nlines = 0;
+    Dwarf_Lines *lines = NULL;
+    func_info* last_function = NULL;
+    Dwarf_Addr last_function_entrypc = 0;
+    Dwarf_Addr last_function_highpc = 0;
     int choose_next_line_otherthan = -1;
 
     // XXX: ideally, there would be a dwarf_getfile(line) routine,
@@ -777,36 +976,72 @@ dwflpp
 	Dwarf_Addr addr;
 	Dwarf_Line * line_rec = dwarf_onesrcline(lines, i);
 	dwarf_lineaddr (line_rec, &addr);
-        int this_lineno;
 
+        // If this next line goes past the current function, we have
+        // a probable tail call, and must go back the entrypc
+        if (choose_next_line_otherthan >= 0 &&
+            addr >= last_function_highpc) // NB: >= ; highpc is exclusive
+          {
+            addr = last_function_entrypc; // go back to entrypc
+            Dwarf_Addr addr0 = last_function->prologue_end;
+            if (addr0 != addr)
+              {
+                last_function->prologue_end = addr;
+		if (sess.verbose)
+		  clog << "prologue disagreement (tail call): "
+                       << last_function->name
+		       << " heur0=" << hex << addr0
+		       << " heur1=" << addr << dec
+		       << "\n";
+              }
+	    choose_next_line_otherthan = -1;
+            continue;
+          }
+
+        // If this next line maps to a line number beyond the
+        // first one, but still within the same function, pick it!
+        int this_lineno;
         dwfl_assert ("dwarf_lineno",
                      dwarf_lineno(line_rec, &this_lineno));
 
 	if (choose_next_line_otherthan >= 0 &&
             this_lineno != choose_next_line_otherthan)
 	  {
-	    map<Dwarf_Addr, func_info>::iterator i =
-              funcs.find (last_function_entrypc);
-	    assert (i != funcs.end());
-            Dwarf_Addr addr0 = i->second.prologue_end;
+            Dwarf_Addr addr0 = last_function->prologue_end;
             if (addr0 != addr)
               {
-                i->second.prologue_end = addr;
+                last_function->prologue_end = addr;
 		if (sess.verbose)
-		  clog << "prologue disagreement: " << i->second.name
+		  clog << "prologue disagreement: " << last_function->name
 		       << " heur0=" << hex << addr0
 		       << " heur1=" << addr << dec
-		       << endl;
+		       << "\n";
               }
 	    choose_next_line_otherthan = -1;
+            continue;
 	  }
 
-	map<Dwarf_Addr, func_info>::const_iterator i = funcs.find (addr);
+        // Else ... see if this line record refers to the entrypc of
+        // yet another probe-targeted function.  Stash away its entrypc
+        // etc., for the other branches of this loop to process.
+	map<Dwarf_Addr, func_info>::iterator i = funcs.find (addr);
 	if (i != funcs.end())
           {
             dwfl_assert ("dwarf_lineno",
-                         dwarf_lineno(line_rec, &choose_next_line_otherthan));
+                         dwarf_lineno (line_rec, &choose_next_line_otherthan));
+            assert (choose_next_line_otherthan >= 0);
+            last_function = & i->second;
             last_function_entrypc = addr;
+            dwfl_assert ("dwarf_highpc",
+                         dwarf_highpc (& last_function->die, 
+                                       & last_function_highpc));
+
+            if (sess.verbose)
+              clog << "finding prologue for '"
+                   << last_function->name
+                   << "' entrypc=0x" << hex << addr 
+                   << " highpc=0x" << last_function_highpc
+                   << "\n";
           }
       }
 
@@ -1200,7 +1435,7 @@ dwflpp
       clog << "finding location for local '" << local
 	   << "' near address " << hex << pc
 	   << ", module bias " << module_bias << dec
-	   << endl;
+	   << "\n";
 
     Dwarf_Attribute attr_mem;
     if (dwarf_attr_integrate (&vardie, DW_AT_location, &attr_mem) == NULL)
@@ -1431,7 +1666,18 @@ dwarf_query
 
 struct dwarf_builder: public derived_probe_builder
 {
-  dwarf_builder() {}
+  dwflpp *kern_dw;  
+  dwflpp *user_dw;  
+  dwarf_builder() 
+    : kern_dw(NULL), user_dw(NULL) 
+  {}
+  ~dwarf_builder() 
+  { 
+    if (kern_dw) 
+      delete kern_dw;
+    if (user_dw) 
+      delete user_dw;
+  }
   virtual void build(systemtap_session & sess,
 		     probe * base,
 		     probe_point * location,
@@ -1546,7 +1792,7 @@ dwarf_query::parse_function_spec(string & spec)
       if (sess.verbose)
 	clog << "parsed '" << spec
 	     << "' -> func '" << function
-	     << "'" << endl;
+	     << "'\n";
       return function_alone;
     }
 
@@ -1562,7 +1808,7 @@ dwarf_query::parse_function_spec(string & spec)
 	clog << "parsed '" << spec
 	     << "' -> func '"<< function
 	     << "', file '" << file
-	     << "'" << endl;
+	     << "'\n";
       return function_and_file;
     }
 
@@ -1576,7 +1822,7 @@ dwarf_query::parse_function_spec(string & spec)
 	clog << "parsed '" << spec
 	     << "' -> func '"<< function
 	     << "', file '" << file
-	     << "', line " << line << endl;
+	     << "', line " << line << "\n";
       return function_file_and_line;
     }
   catch (runtime_error & exn)
@@ -1687,7 +1933,7 @@ dwarf_query::blacklisted_p(string const & funcname,
 	  if (sess.verbose)
 	    clog << "skipping function '" << funcname << "' base 0x"
 		 << hex << addr << dec << " is within section '"
-		 << name << "'" << endl;
+		 << name << "'\n";
 	  return true;
 	}
     }
@@ -1722,7 +1968,7 @@ dwarf_query::blacklisted_p(string const & funcname,
 		  if (sess.verbose)
 		    clog << "skipping function '" << funcname << "' base 0x"
 			 << hex << addr << dec << " is within section '"
-			 << name << "'" << endl;
+			 << name << "'\n";
 		  return true;
 		}
 	    }
@@ -1736,6 +1982,7 @@ dwarf_query::blacklisted_p(string const & funcname,
   if (filename == 0) filename = ""; // possibly 0
   string filename_s = filename; // is passed as const char*
   if (funcname == "do_IRQ" ||
+      funcname == "notifier_call_chain" ||
       filename_s == "kernel/kprobes.c" ||
       0 == fnmatch ("arch/*/kernel/kprobes.c", filename, 0) ||
       (has_return && (funcname == "sys_execve" || // bug #1345
@@ -1744,7 +1991,7 @@ dwarf_query::blacklisted_p(string const & funcname,
     {
       if (sess.verbose)
         clog << "skipping function '" << funcname << "' file '"
-             << filename << "' is blacklisted" << endl;
+             << filename << "' is blacklisted\n";
       return true;
     }
 
@@ -1864,7 +2111,7 @@ query_inline_instance_info (Dwarf_Addr entrypc,
 	  if (q->sess.verbose)
 	    clog << "querying entrypc "
 		 << hex << entrypc << dec
-		 << " of instance of inline '" << ii.name << "'" << endl;
+		 << " of instance of inline '" << ii.name << "'\n";
 	  query_statement (ii.name, ii.decl_file, ii.decl_line,
 			   &ii.die, entrypc, q);
 	}
@@ -1886,9 +2133,6 @@ query_func_info (Dwarf_Addr entrypc,
 	{
 	  // NB. dwarf_derived_probe::emit_registrations will emit a
 	  // kretprobe based on the entrypc in this case.
-	  if (q->sess.verbose)
-	    clog << "querying entrypc of function '"
-		 << fi.name << "' for return probe" << endl;
 	  query_statement (fi.name, fi.decl_file, fi.decl_line,
 			   &fi.die, entrypc, q);
 	}
@@ -1897,21 +2141,13 @@ query_func_info (Dwarf_Addr entrypc,
 #ifdef __ia64__
 	// In IA64 platform function probe point is set at its
 	// entry point rather than prologue end pointer
-	if (q->sess.verbose)
-	   clog << "querying entrypc of function '"
-		<< fi.name << "'" << endl;
 	   query_statement (fi.name, fi.decl_file, fi.decl_line,
 		&fi.die, entrypc, q);
 
 #else
-	  if (q->sess.verbose)
-	    clog << "querying prologue-end of function '"
-		 << fi.name << "'" << endl;
-
 	  if (fi.prologue_end == 0)
 	    throw semantic_error("could not find prologue-end "
 				 "for probed function '" + fi.name + "'");
-
 	  query_statement (fi.name, fi.decl_file, fi.decl_line,
 			   &fi.die, fi.prologue_end, q);
 #endif
@@ -1938,7 +2174,7 @@ query_srcfile_line (Dwarf_Line * line, void * arg)
       if (q->dw.die_has_pc (&(i->second.die), addr))
 	{
 	  if (q->sess.verbose)
-	    clog << "function DIE lands on srcfile" << endl;
+	    clog << "function DIE lands on srcfile\n";
 	  if (q->has_statement_str)
 	    query_statement (i->second.name, i->second.decl_file,
 			     q->line, NULL, addr, q);
@@ -1954,7 +2190,7 @@ query_srcfile_line (Dwarf_Line * line, void * arg)
       if (q->dw.die_has_pc (&(i->second.die), addr))
 	{
 	  if (q->sess.verbose)
-	    clog << "inline instance DIE lands on srcfile" << endl;
+	    clog << "inline instance DIE lands on srcfile\n";
 	  if (q->has_statement_str)
 	    query_statement (i->second.name, i->second.decl_file,
 			     q->line, NULL, addr, q);
@@ -1977,7 +2213,7 @@ query_dwarf_inline_instance (Dwarf_Die * die, void * arg)
       bool record_this_inline = false;
 
       if (q->sess.verbose)
-	clog << "examining inline instance of " << q->dw.function_name << endl;
+	clog << "examining inline instance of " << q->dw.function_name << "\n";
 
       if (q->has_inline_str || q->has_statement_str)
 	record_this_inline = true;
@@ -1996,7 +2232,8 @@ query_dwarf_inline_instance (Dwarf_Die * die, void * arg)
       if (record_this_inline)
 	{
 	  if (q->sess.verbose)
-	    clog << "selected inline instance of " << q->dw.function_name << endl;
+	    clog << "selected inline instance of " << q->dw.function_name
+                 << "\n";
 
 	  Dwarf_Addr entrypc;
 	  if (q->dw.die_entrypc (die, &entrypc))
@@ -2043,7 +2280,8 @@ query_dwarf_func (Dwarf_Die * func, void * arg)
 	      || q->has_inline_num))
 	{
 	  if (q->sess.verbose)
-	    clog << "checking instances of inline " << q->dw.function_name << endl;
+	    clog << "checking instances of inline " << q->dw.function_name
+                 << "\n";
 	  q->dw.iterate_over_inline_instances (query_dwarf_inline_instance, arg);
 	}
       else if (!q->dw.func_is_inline ())
@@ -2072,7 +2310,7 @@ query_dwarf_func (Dwarf_Die * func, void * arg)
 	  if (record_this_function)
 	    {
 	      if (q->sess.verbose)
-		clog << "selected function " << q->dw.function_name << endl;
+		clog << "selected function " << q->dw.function_name << "\n";
 
 	      Dwarf_Addr entrypc;
 	      if (q->dw.function_entrypc (&entrypc))
@@ -2109,7 +2347,7 @@ query_cu (Dwarf_Die * cudie, void * arg)
 
       if (false && q->sess.verbose)
         clog << "focused on CU '" << q->dw.cu_name
-             << "', in module '" << q->dw.module_name << "'" << endl;
+             << "', in module '" << q->dw.module_name << "'\n";
 
       if (q->has_statement_str
 	  || q->has_inline_str || q->has_inline_num
@@ -2220,7 +2458,7 @@ query_module (Dwfl_Module *mod __attribute__ ((unused)),
   dwarf_query * q = static_cast<dwarf_query *>(arg);
 
   try
-    {
+    {      
       q->dw.focus_on_module(mod);
 
       // If we have enough information in the pattern to skip a module and
@@ -2236,7 +2474,7 @@ query_module (Dwfl_Module *mod __attribute__ ((unused)),
 	clog << "focused on module '" << q->dw.module_name
 	     << "' = [" << hex << q->dw.module_start
 	     << "-" << q->dw.module_end
-	     << ", bias " << q->dw.module_bias << "]" << dec << endl;
+	     << ", bias " << q->dw.module_bias << "]" << dec << "\n";
 
       if (q->has_inline_num || q->has_function_num || q->has_statement_num)
         {
@@ -2266,6 +2504,8 @@ query_module (Dwfl_Module *mod __attribute__ ((unused)),
           // Otherwise if we have a function("foo") or statement("foo")
           // specifier, we have to scan over all the CUs looking for
           // the function(s) in question
+	  
+	  q->dw.limit_search_to_function_pattern(q->function);
 
           assert(q->has_function_str || q->has_inline_str || q->has_statement_str);
           q->dw.iterate_over_cus(&query_cu, q);
@@ -2370,17 +2610,21 @@ var_expanding_copy_visitor::visit_target_symbol (target_symbol *e)
 
   // Synthesize a function.
   functiondecl *fdecl = new functiondecl;
+  fdecl->tok = e->tok;
   embeddedcode *ec = new embeddedcode;
   ec->tok = e->tok;
   bool lvalue = is_active_lvalue(e);
 
   if (lvalue && !q.sess.guru_mode)
-    throw semantic_error("Illegal target variable access", e->tok);
+    throw semantic_error("write to target variable not permitted", e->tok);
 
   string fname = (string(lvalue ? "set" : "get")
 		  + "_" + e->base_name.substr(1)
 		  + "_" + lex_cast<string>(tick++));
 
+  if (q.has_return)
+    throw semantic_error ("target variables not available to .return probes");
+
   try
     {
       ec->code = q.dw.literal_stmt_for_local (scope_die,
@@ -2724,20 +2968,32 @@ dwarf_derived_probe::emit_deregistrations (translator_output* o, unsigned proben
 {
   o->newline() << "{";
   o->newline(1) << "int i;";
-  o->newline() << "for (i = 0; i < " << probe_points.size() << "; i++)";
+  o->newline() << "for (i = 0; i < " << probe_points.size() << "; i++) {";
   string probe_name = struct_kprobe_array_name(probenum) + "[i]";
   o->indent(1);
   if (has_return)
     {
       o->newline() << "#ifdef ARCH_SUPPORTS_KRETPROBES";
+      o->newline() << "atomic_add ("
+                   << probe_name << ".kp.nmissed,"
+                   << "& skipped_count);";
+      o->newline() << "atomic_add ("
+                   << probe_name << ".nmissed,"
+                   << "& skipped_count);";
       o->newline() << "unregister_kretprobe (&(" << probe_name << "));";
       o->newline() << "#else";
       o->newline() << ";";
       o->newline() << "#endif";
     }
   else
-    o->newline() << "unregister_kprobe (&(" << probe_name << "));";
-  o->indent(-1);
+    {
+      o->newline() << "atomic_add ("
+                   << probe_name << ".nmissed,"
+                   << "& skipped_count);";
+      o->newline() << "unregister_kprobe (&(" << probe_name << "));";
+    }
+
+  o->newline(-1) << "}";
   o->newline(-1) << "}";
 }
 
@@ -2848,7 +3104,7 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o,
   else
     o->line() << "struct kprobe *probe_instance";
   o->line() << ", struct pt_regs *regs) {";
-  o->newline(1) << "struct context* c = per_cpu_ptr (contexts, smp_processor_id());";
+  o->indent(1);
 
   // Calculate the name of the current probe by finding its index in the probe array.
   if (has_return)
@@ -2859,37 +3115,16 @@ dwarf_derived_probe::emit_probe_entries (translator_output* o,
     o->newline() << "const char* probe_point = "
 		 << string_array
 		 << "[ (probe_instance - &(" << probe_array << "[0]))];";
-
-  // A precondition for running a probe handler is that we're in RUNNING
-  // state (not ERROR), and that no one else is already using this context.
-  o->newline() << "if (atomic_read (&session_state) != STAP_SESSION_RUNNING)";
-  o->newline(1) << "return 0;";
-  o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
-  o->newline(1) << "_stp_warn (\"probe reentrancy (%s vs %s)\\n\", "
-		    << "c->probe_point, probe_point);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline() << "atomic_dec (&c->busy);";
-  o->newline() << "return 0;";
-  o->newline(-1) << "}";
-  o->newline();
-
-  o->newline() << "c->last_error = 0;";
-  o->newline() << "c->probe_point = probe_point;";
-  o->newline() << "c->nesting = 0;";
+  emit_probe_prologue (o, "STAP_SESSION_RUNNING");
   o->newline() << "c->regs = regs;";
-  o->newline() << "c->actioncount = 0;";
 
   // NB: locals are initialized by probe function itself
   o->newline() << "probe_" << probenum << " (c);";
 
-  o->newline() << "if (c->last_error && c->last_error[0]) {";
-  o->newline(1) << "_stp_error (\"%s near %s\", c->last_error, c->last_stmt);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline(-1) << "}";
+  emit_probe_epilogue (o);
 
-  o->newline() << "atomic_dec (& c->busy);";
   o->newline() << "return 0;";
-  o->newline(-1) << "}" << endl;
+  o->newline(-1) << "}\n";
   if (has_return)
     o->newline() << "#endif /* ARCH_SUPPORTS_KRETPROBES */";
 
@@ -2904,11 +3139,36 @@ dwarf_builder::build(systemtap_session & sess,
 		     std::map<std::string, literal *> const & parameters,
 		     vector<derived_probe *> & finished_results)
 {
+  dwflpp *dw = NULL;
 
-  dwflpp dw(sess);
-  dwarf_query q(sess, base, location, dw, parameters, finished_results);
+  string dummy;
+  bool has_kernel = dwarf_query::has_null_param(parameters, TOK_KERNEL);
+  bool has_module = dwarf_query::get_string_param(parameters, TOK_MODULE, dummy);
+
+  if (has_kernel || has_module)
+    {
+      if (!kern_dw)
+	{
+	  kern_dw = new dwflpp(sess);
+	  assert(kern_dw);
+	  kern_dw->setup(true);
+	}
+      dw = kern_dw;
+    }
+  else
+    {
+      if (!user_dw)
+	{
+	  user_dw = new dwflpp(sess);
+	  assert(user_dw);
+	  user_dw->setup(false);
+	}
+      dw = user_dw;
+    }
+  assert(dw);
+
+  dwarf_query q(sess, base, location, *dw, parameters, finished_results);
 
-  dw.setup(q.has_kernel || q.has_module);
 
   if (q.has_kernel &&
       (q.has_function_num || q.has_inline_num || q.has_statement_num))
@@ -2924,8 +3184,8 @@ dwarf_builder::build(systemtap_session & sess,
 	a = q.inline_num_val;
       else
 	a = q.statement_num_val;
-      dw.focus_on_module_containing_global_address(a);
-      dw.query_cu_containing_global_address(a, &q);
+      dw->focus_on_module_containing_global_address(a);
+      dw->query_cu_containing_global_address(a, &q);
     }
   else
     {
@@ -2939,12 +3199,12 @@ dwarf_builder::build(systemtap_session & sess,
       if (q.has_kernel)
 	{
 	  int flag = 0;
-	  dw.iterate_over_modules(&query_kernel_exists, &flag);
+	  dw->iterate_over_modules(&query_kernel_exists, &flag);
 	  if (! flag)
 	    throw semantic_error ("cannot find kernel debuginfo");
 	}
 
-      dw.iterate_over_modules(&query_module, &q);
+      dw->iterate_over_modules(&query_module, &q);
     }
 }
 
@@ -3014,26 +3274,12 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
   o->newline() << "static struct timer_list timer_" << j << ";";
 
   o->newline() << "void enter_" << j << " (unsigned long val) {";
-  o->newline(1) << "struct context* c = per_cpu_ptr (contexts, smp_processor_id());";
+  o->indent(1);
   o->newline() << "const char* probe_point = "
 	       << lex_cast_qstring(*locations[0]) << ";";
+  emit_probe_prologue (o, "STAP_SESSION_RUNNING");
+
   o->newline() << "(void) val;";
-
-  // A precondition for running a probe handler is that we're in
-  // RUNNING state (not ERROR), and that no one else is already using
-  // this context.
-  o->newline() << "if (atomic_read (&session_state) != STAP_SESSION_RUNNING)";
-  o->newline(1) << "return;";
-
-  o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
-  o->newline(1) << "_stp_warn (\"probe reentrancy (%s vs %s)\\n\", "
-		<< "c->probe_point, probe_point);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline() << "atomic_dec (&c->busy);";
-  o->newline() << "return;";
-  o->newline(-1) << "}";
-  o->newline();
-
   o->newline() << "mod_timer (& timer_" << j << ", jiffies + ";
   if (time_is_msecs)
     o->line() << "msecs_to_jiffies(";
@@ -3044,22 +3290,11 @@ timer_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
     o->line() << ")";
   o->line() << ");";
 
-  o->newline() << "c->probe_point = probe_point;";
-  o->newline() << "c->last_error = 0;";
-  o->newline() << "c->nesting = 0;";
-  o->newline() << "c->regs = 0;";
-  o->newline() << "c->actioncount = 0;";
-
   // NB: locals are initialized by probe function itself
   o->newline() << "probe_" << j << " (c);";
 
-  o->newline() << "if (c->last_error && c->last_error[0]) {";
-  o->newline(1) << "_stp_error (\"%s near %s\", c->last_error, c->last_stmt);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline(-1) << "}";
-
-  o->newline() << "atomic_dec (&c->busy);";
-  o->newline(-1) << "}" << endl;
+  emit_probe_epilogue (o);
+  o->newline(-1) << "}\n";
 }
 
 
@@ -3098,7 +3333,11 @@ struct timer_builder: public derived_probe_builder
 
 struct profile_derived_probe: public derived_probe
 {
-  profile_derived_probe (probe* p, probe_point* l): derived_probe(p, l) {}
+  // kernels < 2.6.10: use register_profile_notifier API
+  // kernels >= 2.6.10: use register_timer_hook API
+  bool using_rpn;
+
+  profile_derived_probe (systemtap_session &s, probe* p, probe_point* l);
 
   virtual void emit_registrations (translator_output * o, unsigned i);
   virtual void emit_deregistrations (translator_output * o, unsigned i);
@@ -3106,94 +3345,75 @@ struct profile_derived_probe: public derived_probe
 };
 
 
+profile_derived_probe::profile_derived_probe (systemtap_session &s, probe* p, probe_point* l):
+  derived_probe(p, l)
+{
+  string target_kernel_v;
+
+  // cut off any release code suffix
+  string::size_type dash_rindex = s.kernel_release.rfind ('-');
+  if (dash_rindex > 0 && dash_rindex != string::npos)
+    target_kernel_v = s.kernel_release.substr (0, dash_rindex);
+  else
+    target_kernel_v = s.kernel_release;
+
+  using_rpn = (strverscmp(target_kernel_v.c_str(), "2.6.10") < 0);
+}
+
+
 void
 profile_derived_probe::emit_registrations (translator_output* o, unsigned j)
 {
-  // kernels < 2.6.10: use register_profile_notifier API
-  o->newline() << "#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)";
-  o->newline() << "rc = register_profile_notifier(&profile_" << j << ");";
-
-  // kernels >= 2.6.10: use register_timer_hook API
-  o->newline() << "#else";
-  o->newline() << "rc = register_timer_hook(enter_" << j << ");";
-
-  o->newline() << "#endif";
+  if (using_rpn)
+    o->newline() << "rc = register_profile_notifier(&profile_" << j << ");";
+  else
+    o->newline() << "rc = register_timer_hook(enter_" << j << ");";
 }
 
 
 void
 profile_derived_probe::emit_deregistrations (translator_output* o, unsigned j)
 {
-  // kernels < 2.6.10: use register_profile_notifier API
-  o->newline() << "#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)";
-  o->newline() << "unregister_profile_notifier(&profile_" << j << ");";
-
-  // kernels >= 2.6.10: use register_timer_hook API
-  o->newline() << "#else";
-  o->newline() << "unregister_timer_hook(enter_" << j << ");";
-
-  o->newline() << "#endif";
+  if (using_rpn)
+    o->newline() << "unregister_profile_notifier(&profile_" << j << ");";
+  else
+    o->newline() << "unregister_timer_hook(enter_" << j << ");";
 }
 
 
 void
 profile_derived_probe::emit_probe_entries (translator_output* o, unsigned j)
 {
-  // kernels < 2.6.10: use register_profile_notifier API
-  o->newline() << "#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 10)";
-  o->newline() << "static int enter_" << j
-	       << " (struct notifier_block *self, unsigned long val, void *data);";
-  o->newline() << "static struct notifier_block profile_" << j << " = {";
-  o->newline(1) << ".notifier_call = enter_" << j << ",";
-  o->newline(-1) << "};";
-  o->newline() << "int enter_" << j
-	       << " (struct notifier_block *self, unsigned long val, void *data) {";
-  o->newline(1) << "struct pt_regs *regs = (struct pt_regs *)data;";
+  if (using_rpn) {
+    o->newline() << "static int enter_" << j
+                 << " (struct notifier_block *self, unsigned long val, void *data);";
+    o->newline() << "static struct notifier_block profile_" << j << " = {";
+    o->newline(1) << ".notifier_call = enter_" << j << ",";
+    o->newline(-1) << "};";
+    o->newline() << "int enter_" << j
+                 << " (struct notifier_block *self, unsigned long val, void *data) {";
+    o->newline(1) << "struct pt_regs *regs = (struct pt_regs *)data;";
+    o->indent(-1);
+  } else {
+    o->newline() << "static int enter_" << j << " (struct pt_regs *regs);";
+    o->newline() << "int enter_" << j << " (struct pt_regs *regs) {";
+  }
 
-  // kernels >= 2.6.10: use register_timer_hook API
-  o->newline(-1) << "#else";
-  o->newline() << "static int enter_" << j << " (struct pt_regs *regs);";
-  o->newline() << "int enter_" << j << " (struct pt_regs *regs) {";
-  o->newline(1) << "unsigned long val = 0;";
-  o->newline(-1) << "#endif";
-
-  o->newline(1) << "struct context* c = per_cpu_ptr (contexts, smp_processor_id());";
+  o->indent(1);
   o->newline() << "const char* probe_point = "
 	       << lex_cast_qstring(*locations[0]) << ";";
+  emit_probe_prologue (o, "STAP_SESSION_RUNNING");
 
-  // A precondition for running a probe handler is that we're in
-  // RUNNING state (not ERROR), and that no one else is already using
-  // this context.
-  o->newline() << "(void) val;";
-  o->newline() << "if (atomic_read (&session_state) != STAP_SESSION_RUNNING)";
-  o->newline(1) << "return 0;";
+  if (using_rpn) {
+    o->newline() << "(void) self;";
+    o->newline() << "(void) val;";
+  }
 
-  o->newline(-1) << "if (atomic_inc_return (&c->busy) != 1) {";
-  o->newline(1) << "_stp_warn (\"probe reentrancy (%s vs %s)\\n\", "
-		<< "c->probe_point, probe_point);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline() << "atomic_dec (&c->busy);";
-  o->newline() << "return 0;";
-  o->newline(-1) << "}";
-  o->newline();
-
-  o->newline() << "c->probe_point = probe_point;";
-  o->newline() << "c->last_error = 0;";
-  o->newline() << "c->nesting = 0;";
-  o->newline() << "c->regs = regs;";
-  o->newline() << "c->actioncount = 0;";
-
-  // NB: locals are initialized by probe function itself
   o->newline() << "probe_" << j << " (c);";
 
-  o->newline() << "if (c->last_error && c->last_error[0]) {";
-  o->newline(1) << "_stp_error (\"%s near %s\", c->last_error, c->last_stmt);";
-  o->newline() << "atomic_set (& session_state, STAP_SESSION_ERROR);";
-  o->newline(-1) << "}";
-
-  o->newline() << "atomic_dec (&c->busy);";
+  emit_probe_epilogue (o);
   o->newline() << "return 0;";
-  o->newline(-1) << "}" << endl;
+  o->newline(-1) << "}\n";
 }
 
 
@@ -3206,7 +3426,7 @@ struct profile_builder: public derived_probe_builder
 		     std::map<std::string, literal *> const & parameters,
 		     vector<derived_probe *> & finished_results)
   {
-    finished_results.push_back(new profile_derived_probe(base, location));
+    finished_results.push_back(new profile_derived_probe(sess, base, location));
   }
 };
 
